Attribute VB_Name = "Physics"
'
'                 P H Y S I C S
'

'Important

Option Explicit

' experimental object: used to store a grid of environmental
' features
Public envgridpres As Boolean
Public nlink As Single         ' links physics constants
Public plink As Single
Public klink As Single
Public mlink As Single

Public Const smudgefactor As Single = 50 'just to keep the bots more likely to stay visible

Public Function NetForces(n As Integer)
  Dim mag As Single
  Dim sign As Integer
  Dim staticV As vector
  
  'The physics engine breaks apart if bot masses are less than about .1
  
  With rob(n)
  
  CalcMass n
  AddedMass n 'not really a force, but does effect end net acceleration
  
  PlanetEaters n
  TieHooke n
  FrictionForces n
  SphereDragForces n
  'TieDragForces n
  
  BrownianForces n
  BouyancyForces n
  GravityForces n
  GateForces n
  VoluntaryForces n
  
  bordercolls n
  Colls2 n
  
  If .ImpulseStatic > 0 Then
    staticV = VectorScalar(VectorUnit(.ImpulseInd), .ImpulseStatic)
    If VectorMagnitudeSquare(staticV) > VectorMagnitudeSquare(.ImpulseInd) Then
      .ImpulseInd = VectorSub(.ImpulseInd, staticV)
    End If
  End If
  
  .ImpulseInd = VectorSub(.ImpulseInd, .ImpulseRes)
  End With
End Function

Private Sub CalcMass(n As Integer)
  With rob(n)
  .mass = (.BODY / 1000) + (.Shell / 200) 'set value for mass
  If .mass < 0.1 Then .mass = 0.1 'stops the Euler integration from wigging out too badly.
  
  End With
End Sub

Public Sub AddedMass(n As Integer)
  'added mass is a simple enough concept.
  'To move an object through a liquid, you must also move
  'that liquid out of the way.
  
  Const fourthirdspi As Single = 4.18879
  
  Const AddedMassCoefficientForASphere As Single = 0.5
  
  With rob(n)
    .AddedMass = AddedMassCoefficientForASphere * _
      SimOpts.Density * fourthirdspi * .radius * .radius * .radius
  End With
End Sub

Public Sub FrictionForces(n As Integer)
  Dim Impulse As Single
  Dim mag As Single
  With rob(n)
  
  If SimOpts.Zgravity = 0 Then Exit Sub
  
  If .vel.x = 0 And .vel.y = 0 Then 'is there a vector way to do this?
    .ImpulseStatic = .mass * SimOpts.Zgravity * SimOpts.CoefficientStatic ' * 1 cycle (timestep = 1)
  Else
    Impulse = .mass * SimOpts.Zgravity * SimOpts.CoefficientKinetic ' * 1 cycle (timestep = 1)
    .ImpulseRes = VectorScalar(VectorUnit(.vel), -Impulse) 'kinetic friction points in opposite direction of velocity
  End If
  End With
End Sub

Public Sub BrownianForces(n As Integer)
  If SimOpts.PhysBrown = 0 Then Exit Sub
  Dim Impulse As Single
  Dim RandomAngle As Single
  With rob(n)
  
    Impulse = Gauss(SimOpts.PhysBrown / 2)
    RandomAngle = fRnd(1, 1256) / 200
    .ImpulseInd = VectorAdd(.ImpulseInd, VectorSet(Cos(RandomAngle) * Impulse, Sin(RandomAngle) * Impulse))
  End With
End Sub

Public Sub SphereDragForces(n As Integer) 'for bots
  Dim Impulse As Single
  Dim ImpulseVector  As vector
  Dim mag As Single
  
  If rob(n).vel.x = 0 And rob(n).vel.y = 0 Then Exit Sub
  
  mag = VectorMagnitude(rob(n).vel)
  
  Impulse = 0.5 * SphereCd(mag, rob(n).radius) * _
    SimOpts.Density * mag * mag * (PI * rob(n).radius ^ 2)
  ImpulseVector = VectorScalar(VectorUnit(rob(n).vel), Impulse)
  rob(n).ImpulseRes = VectorAdd(rob(n).ImpulseRes, ImpulseVector)
End Sub

Public Sub TieDragForces(n As Integer) 'for ties
'calculate drag on the ties as if the ties are cylinders
'radius of the tie should be stored in tie array
Dim a As Long

For a = 0 To 10
  If rob(n).Ties(a).pnt > 0 Then
    'If rob(n).Ties(a).pnt > n Then TieDrag2 n, rob(n).Ties(a).pnt
    TieDrag n, rob(n).Ties(a).pnt
  End If
Next a

End Sub

Public Sub TieDrag3(n1 As Integer, N2 As Integer)
  Dim pos As vector
  Dim a As Single, b As Single, c As Single
  Dim TorqueScalar
  
  pos = VectorSub(rob(N2).pos, rob(n1).pos)
  
  a = Cross(rob(n1).vel, pos)
  b = Cross(rob(N2).vel, pos)
  c = (a + b) * 0.5 * 0.0001
  
  'pretend Cd is always 1
  pos = VectorUnit(pos)
  pos = VectorSet(-pos.y, pos.x)
  rob(n1).ImpulseRes = VectorSub(rob(n1).ImpulseRes, VectorScalar(pos, Abs(c)))
  
End Sub

Public Sub TieDrag2(ByVal n1 As Long, ByVal N2 As Long)
  Dim pos As vector
  Dim a As Single, b As Single, c As Single
  Dim invlength As Single
  Dim ForceScalar As Single
  
  If rob(n1).mass = 0 Or rob(N2).mass = 0 Then Exit Sub
  
  'cooperative or independant\
  pos = VectorSub(rob(n1).pos, rob(N2).pos)
  invlength = VectorInvMagnitude(pos)
  
  'a and b are the two cross velocities
  a = Cross(rob(n1).vel, pos)
  b = Cross(rob(N2).vel, pos)
  
  c = (a + b) * 0.5 * invlength 'the average cross velocity
  'use c to find the Cd
  
  Dim asquare As Single
  Dim ab As Single
  Dim bsquare As Single
  Dim BigB As Single
  Dim BigA As Single
  Const TieRadius As Single = 30
  
  asquare = a * a
  ab = a * b
  bsquare = b * b
  BigB = rob(N2).mass / (rob(n1).mass + rob(N2).mass)
  BigA = SimOpts.Density * TieRadius * CylinderCd(c, TieRadius) * 0.083333333333 * Sgn(c)
  
  ForceScalar = BigA * invlength * (4 * BigB * _
    (asquare - 5 * ab + bsquare) + asquare + 2 * ab + 3 * bsquare)
  'divide the above be either B or 1-B (depending on which robot we're
  'applying forces to) and multiply by the orthogonal unit component to pos
  pos = VectorScalar(pos, invlength)
  pos = VectorSet(-pos.y, pos.x)
  
  rob(n1).ImpulseRes = VectorAdd(rob(n1).ImpulseRes, _
    VectorScalar(pos, ForceScalar * 0.5 / BigB))
  'rob(N2).ForceRes = VectorAdd(rob(N2).ForceRes, _
  '  VectorScalar(pos, ForceScalar * 0.5 / (1 - BigB)))
End Sub

Public Sub TieDrag(n1 As Integer, N2 As Integer)
'Simple method:

'v1 = my velocity
'm1 = my mass
'p1 = my position
'v2 = other 's velocity
'm2 = other mass
'p2 = other position
'
'
'1.  Find unit vector for tie = u = (p2 - p1) / length(p2-p1)
'2.  a = v1 cross u  |  a and b are the cross velocities, that is, the velocity
'    b = vc cross u  |  perpindicular to the movement of the tie,
'                    |  which is the direction that causes drag
'    v(d) = a + (b-a)/length * d where d is distance from a
'3.  Force on either bots is: A/12*length^2(a^2+2ab+3b^2) in a direction perpindicular to u.
'A = density * radius * Cd
'4.  Find force of drag per unit length using c as the velocity <--- this is
'     the assumption that drag and velocity are linearly related.
'     For turbulent flows, using the average of two velocities is incorrect.
'     In the future, this should be solved.  This requires integrating the
'     Cd and drag force equations for length (with velocity depending linearly
'     on distance from either Mc or m1 of course) between 0 and distance from
'     m1 to Mc.
'5.  The torque (we just add it to resistive forces) applied to "me" =
'     Drag Force per length / 2, since it's the area of the triangle
'     formed by length and dragforces at Mc and m1, divided by the length
'     since we 're applying them all to m1.

  Dim u As vector, vc As vector, a As Single, b As Single, c As Single
  Dim Aconstant As Single
  Dim DragScalar As Single
  Dim Drag As vector
  Dim radius As Single
  Dim length As Single
  
  '1.  Find unit vector
  u = VectorSub(rob(N2).pos, rob(n1).pos)
  length = VectorMagnitude(u)
  u = VectorUnit(u)
  
  a = Cross(rob(n1).vel, u)
  b = Cross(rob(N2).vel, u)
  c = (a * a + 2 * a * b + 3 * b * b)
  '4.  Find drag using c:
  
  '1.7 is a good radius.  What?  Whhhaaatttt?  It is.
  'okay, it's because that's what 10 body, at 905 twips^3 each,
  'stretched into a cylinder with a length of 1000 twips would be
    
  radius = Sqr(9050 / length / PI)
  Aconstant = radius * SimOpts.Density * 1
    
  DragScalar = Aconstant * length * length / 12 * c
  Drag = VectorScalar(VectorSet(-u.y, u.x), DragScalar * 0.5)
  
  '5:  apply drag to bot
  'not working right :/
  'rob(n1).ForceRes = VectorAdd(rob(n1).ForceRes, Drag)
End Sub

Public Function SphereCd(ByVal velocitymagnitude As Single, ByVal radius As Single) As Single
  'computes the coeficient of drag for a spehre given the unit reynolds in simopts
  'totally ripped from an online drag calculator.  So sue me.
  
  With SimOpts
  
  Dim Reynolds As Single, y11 As Single, y12 As Single, y13 As Single, y1 As Single, y2 As Single, alpha As Single
  If .Viscosity = 0 Then Exit Function
  Reynolds = radius * 2 * velocitymagnitude * .Density / .Viscosity
  If Reynolds = 0 Then
    SphereCd = 0
    Exit Function
  End If
  
  y11 = 24 / (3 * 10 ^ 5)
  y12 = 6 / (1 + Sqr(3 * 10 ^ 5))
  y13 = 0.4
  
  y1 = y11 + y12 + y13
  y2 = 0.09
  
  alpha = (y2 - y1) * 50000 ^ -2

  If Reynolds < 3 * 10 ^ 5 Then
    SphereCd = 24 / Reynolds + 6 / (1 + Sqr(Reynolds)) + 0.4
  ElseIf Reynolds < 3.5 * 10 ^ 5 Then
    SphereCd = alpha * (Reynolds - (3 * 10 ^ 5)) ^ 2 + y1
  ElseIf Reynolds < 6 * 10 ^ 5 Then
    SphereCd = 0.09
  ElseIf Reynolds < 4 * 10 ^ 6 Then
    SphereCd = (Reynolds / (6 * 10 ^ 5)) ^ 0.55 * y2
  Else
    SphereCd = 0.255
  End If
  End With
End Function

Public Function CylinderCd(ByVal velocitymagnitude As Single, ByVal radius As Single) As Single
  Dim sign As Single
  
  With SimOpts
  
  Const alpha As Single = -3.6444444444444E-11
  
  If velocitymagnitude < 0 Then
    sign = -1#
    velocitymagnitude = -velocitymagnitude
  Else
    sign = 1#
  End If
  
  If .Viscosity = 0 Then
    CylinderCd = 0
    Exit Function
  End If
  
  Dim Reynolds As Single
  Reynolds = radius * 2 * velocitymagnitude * .Density / .Viscosity
  
  Select Case Reynolds
    Case 0
      CylinderCd = 0
    Case Is < 1
      CylinderCd = (8 * PI) / (Reynolds * (Log(8 / Reynolds) - 0.077216))
    Case Is < 100000#
      CylinderCd = 1 + 10 / Reynolds ^ (2 / 3)
    Case Is < 250000#
      CylinderCd = alpha * (Reynolds - 100000) ^ 2 + 1#
    Case Is < 600000#
      CylinderCd = 0.18
    Case Is < 4000000
      CylinderCd = 0.18 * (Reynolds / 600000) ^ 0.63
    Case Is >= 4000000
      CylinderCd = 0.6
  End Select
  End With
End Function

Public Sub BouyancyForces(n As Integer)
  Dim Impulse As Single
  With rob(n)
    If SimOpts.Ygravity = 0 Then Exit Sub
    
    Impulse = -SimOpts.Density * .radius ^ 3 * 4 / 3 * PI * SimOpts.Ygravity
    .ImpulseInd = VectorAdd(.ImpulseInd, VectorSet(0, Impulse))
  End With
End Sub

' a little repulsion field around the download gate, just to
' avoid blocking the way to incoming robots
Public Sub GateForces(n As Integer)
If IntOpts.Active = False Then Exit Sub
  Dim xr As Single
  Dim XCenter As Single
  Dim YCenter As Single
  Dim radius As Single
  Dim dist As Single
  Dim spx As Single
  Dim spy As Single
  Dim k As node
  
  radius = IntOpts.RUpload / 2
  
  xr = IntOpts.XSpawn + IntOpts.RUpload
  
  XCenter = IntOpts.XSpawn + radius
  YCenter = IntOpts.YSpawn + radius
  
  radius = radius * radius * 2.8 '* 2.8 for the edges of the square
  dist = (rob(n).pos.x - XCenter) ^ 2 + (rob(n).pos.y - YCenter) ^ 2
  
  If dist < radius Then
    spx = -((radius - dist) / radius ^ 2) * 200
    spy = (YCenter - rob(n).pos.y) * spx
    spx = (XCenter - rob(n).pos.x) * spx
    rob(n).ImpulseInd = VectorSet(spx, spy)
  End If
End Sub

Public Sub GravityForces(n As Integer)
  With rob(n)
    .ImpulseInd = VectorAdd(.ImpulseInd, VectorSet(0, SimOpts.Ygravity * .mass))
  End With
End Sub

Public Sub VoluntaryForces(n As Integer)
  'calculates new acceleration and energy values from robot's
  '.up/.dn/.sx/.dx vars
  Dim EnergyCost As Single
  Dim NewAccel As vector
  Dim dir As vector
  
  With rob(n)
    'corpses are dead, they don't move around of their own volition
    If .Corpse Then Exit Sub
    
    'yes it's backwards, that's on purpose
    dir = VectorSet(CLng(.mem(dirup)) - CLng(.mem(dirdn)), CLng(.mem(dirsx)) - CLng(.mem(dirdx)))
    NewAccel = VectorSet(Dot(.aimvector, dir), Cross(.aimvector, dir))
    
    'NewAccel is the impulse vector formed by the robot's internal "engine".
    'Impulse is the integral of Force over time.
    
    .ImpulseInd = VectorAdd(.ImpulseInd, VectorScalar(NewAccel, SimOpts.PhysMoving))
    
    EnergyCost = VectorMagnitude(NewAccel) * SimOpts.Costs(MOVECOST)
    If .Multibot Then EnergyCost = EnergyCost / (.Numties + 1)
    .nrg = .nrg - EnergyCost
  End With
End Sub

Public Sub TieHooke(n As Integer, Optional timestep As Single = 0)
  'Handles Hooke forces of a tie.  That is, stretching and shrinking
  'Force = -kx - bv
  'from experiments, k and b should be less than .1 otherwise the forces
  'become too great for a euler modelling (that is, the forces become too large
  'for velocity = velocity + acceleration
  
  'can be made less complex (from O(n^2) to Olog(n) by calculating forces only
  'for robots less than current number and applying that force to both robots
    
  Dim length As Single
  Dim displacement As Single
  Dim Impulse As Single
  Dim k As Integer
  Dim uv As vector
  Dim vy As vector
  
  With rob(n)
        
  For k = 1 To 10
    
    'burn through empty slots in case one is hiding
    While k <= 10 And .Ties(k).pnt <= 0
      k = k + 1
      If k > 10 Then Exit Sub
    Wend
    
    uv = VectorSub(.pos, rob(.Ties(k).pnt).pos)
    length = VectorInvMagnitude(uv)
                
    .Ties(k).b = 0.01
    .Ties(k).k = 0.005
    
    'update ties
    
    'delete tie if length > 1000
    'remember length is inverse squareroot
    If length < 0.001 Then deltie n, .Ties(k).pnt
    
    If .Ties(k).last > 1 Then .Ties(k).last = .Ties(k).last - 1
    If .Ties(k).last = 1 Then deltie n, .Ties(k).pnt
    
    If length <> 0 Then
      uv = VectorScalar(uv, length)
      
      'first -kx
      displacement = .Ties(k).NaturalLength - 1 / length
      Impulse = .Ties(k).k * displacement
      .ImpulseInd = VectorAdd(.ImpulseInd, VectorScalar(uv, Impulse))
      
      'next -bv
      vy = VectorSub(.vel, rob(.Ties(k).pnt).vel)
      Impulse = -Dot(vy, uv) * .Ties(k).b
      .ImpulseInd = VectorAdd(.ImpulseInd, VectorScalar(uv, Impulse))
    End If
  Next k
  End With
End Sub

Public Sub PlanetEaters(n As Integer)
'this way is really, really slow, since we normalize the vector (yuck)
  Dim t As Integer
  Dim Force As Single
  Dim PosDiff As vector
  Dim mag As Single
  
  If Not SimOpts.PlanetEaters Then Exit Sub
  If rob(n).mass = 0 Then Exit Sub
    
  For t = n + 1 To MaxRobs
    If rob(t).mass = 0 Or Not rob(t).Exist Then GoTo Nextiteration
    
    PosDiff = VectorSub(rob(t).pos, rob(n).pos)
    mag = VectorMagnitude(PosDiff)
    If mag = 0 Then GoTo Nextiteration
    
    Force = (SimOpts.PlanetEatersG * rob(n).mass * rob(t).mass) / (mag * mag)
    PosDiff = VectorScalar(PosDiff, 1 / mag)
    'Now set PosDiff to the vector for force along that line
        
    PosDiff = VectorScalar(PosDiff, Force)
    
    rob(n).ImpulseInd = VectorAdd(rob(n).ImpulseInd, PosDiff)
    rob(t).ImpulseInd = VectorSub(rob(t).ImpulseInd, PosDiff)
Nextiteration:
  Next t
End Sub














' calculates angle between (x1,y1) and (x2,y2)
Public Function angle(ByVal x1 As Long, ByVal y1 As Long, ByVal x2 As Long, ByVal y2 As Long) As Single
  Dim an As Single
  Dim dx As Long
  Dim dy As Long
  dx = x2 - x1
  dy = y1 - y2
  If dx = 0 Then
    an = 0
    If dy < 0 Then an = PI / 4 * 3
  Else
    an = Atn(dy / dx)
    If dx < 0 Then
      an = an + PI
    End If
  End If
  angle = an
End Function

' normalizes angle in 0,2pi
Public Function angnorm(ByVal an As Single) As Single
  While an < 0
    an = an + 2 * PI
  Wend
  While an > 2 * PI
    an = an - 2 * PI
  Wend
  angnorm = an
End Function

' calculates difference between two angles
Public Function AngDiff(a1 As Single, a2 As Single) As Single
  Dim r As Single
  r = a1 - a2
  If r > PI Then
    r = -(2 * PI - r)
  End If
  If r < -PI Then
    r = r + 2 * PI
  End If
  AngDiff = r
End Function

'' calculates torque generated by all ties on robots
'Public Sub TieTorque()
'  Dim check As Single
'  Dim anl As Single
'  Dim dlo As Single
'  Dim dx As Long
'  Dim dy As Long
'  Dim dist As Long
'  Dim n As Integer
'  Dim nd As node
'  Dim t As Integer
'  Dim j As Integer
'  Dim mt As Single, mm As Single, m As Single
'  Dim nax As Long, nay As Long
'  Set nd = rlist.firstnode
'  While Not (nd Is rlist.last)
'    t = nd.robn
'    rob(t).mt = 0
'    Set nd = nd.pn
'  Wend
'  Set nd = rlist.firstnode
'  While Not (nd Is rlist.last)
'    j = 1
'    mt = 0
'    t = nd.robn
'    With rob(t)
'      If Not .Corpse Then         'condition added to prevent parsing corpses.
'        If .Numties > 0 Then        'condition added to prevent parsing robots without ties.
'          If .Ties(1).pnt > 0 Then
'            While .Ties(j).pnt > 0
'              If .Ties(j).angreg Then 'if angle is fixed.
'                n = .Ties(j).pnt
'                anl = angle(.x, .y, rob(n).x, rob(n).y) 'angle of tie in euclidian space
'                dlo = AngDiff(anl, .aim) 'difference of angle of tie and direction of robot
'                mm = AngDiff(dlo, .Ties(j).ang + .Ties(j).bend) 'difference of actual angle and requested angle
'                .Ties(j).bend = 0 'reset bend command .tieang
'                m = mm * 0.3
'                dx = rob(n).x - .x
'                dy = .y - rob(n).y
'                dist = Sqr(dx ^ 2 + dy ^ 2)
'                nax = -Sin(anl) * m * dist / 10
'                nay = -Cos(anl) * m * dist / 10
'                'experimental limits to acceleration
'                If Abs(nax) > 100 Then nax = 100 * Sgn(nax)
'                If Abs(nay) > 100 Then nay = 100 * Sgn(nax)
'                rob(n).ax = rob(n).ax - nax
'                rob(n).ay = rob(n).ay - nay
'                .ax = .ax + nax
'                .ay = .ay + nay
'                mt = mt + mm    'in other words mt = mm for 1 tie
'                If t = 10 Then
'                  mt = mt
'                End If
'              End If
'              j = j + 1
'            Wend
'            'If rob(t).absvel > 10 Then rob(30000).absvel = 1000000  'crash inducing line for debugging
'            .ma = angnorm(mt / (j - 1))
'            .aim = angnorm(.aim + .ma)
'            .aimx = Cos(.aim)
'            .aimy = Sin(.aim)
'            '.mem(setaim) = .aim * 200
'          End If
'        End If
'      End If
'    End With
'    Set nd = nd.pn
'  Wend
'End Sub

'' calculates acceleration due to the medium action on the links
'' (used for swimming)
'Public Sub Swimming()
'  Dim vxle As Long
'  Dim vyle As Long
'  Dim nd As node
'  Dim t As Integer, p As Integer
'  Dim j As Byte
'  Dim anle As Single, anve As Single, ancm As Single
'  Dim vea As Long, lle As Long
'  Dim cnorm As Single
'  Dim Fx As Long, Fy As Long
'  Set nd = rlist.firstnode
'  While Not (nd Is rlist.last)
'    t = nd.robn
'    With rob(t)
'      If .Corpse = False And .Numties > 0 Then  'new conditions to prevent parsing corpses and robots without ties.
'        j = 1
'        While .Ties(j).pnt > 0
'          p = .Ties(j).pnt
'          vxle = (.vx + rob(p).vx) / 2  'average x velocity
'          vyle = (.vy + rob(p).vy) / 2  'average y velocity
'          anle = angle(.x, .y, rob(p).x, rob(p).y)  'Angle between robots
'          anve = angle(0, 0, vxle, vyle)            'Angle of vector velocity
'          ancm = anve - anle                        'Combined angle
'          vea = Sqr(vxle ^ 2 + vyle ^ 2)            'velocity along vector
'          lle = Sqr((.x - rob(p).x) ^ 2 + (.y - rob(p).y) ^ 2)  'distance between robots
'          cnorm = Sin(ancm) * vea * lle * SimOpts.PhysSwim  'Swim force
'          'cnorm = Average Cross velocity * distance between bots * PhysSwim
'          Fx = cnorm * Sin(anle) / 800
'          Fy = cnorm * Cos(anle) / 800
'          .ax = .ax + Fx
'          .ay = .ay + Fy
'          rob(p).ax = rob(p).ax + Fx
'          rob(p).ay = rob(p).ay + Fy
'          j = j + 1
'        Wend
'      End If
'    End With
'    Set nd = nd.pn
'  Wend
'End Sub

Public Sub bordercolls(t As Integer, Optional whichside As Integer = 0, Optional whichbottom As Integer = 0)
  'treat the borders as spongy ground
  'that makes you bounce off.
  
  'bottom = -1 for top, 1 for ground
  'side = -1 for left, 1 for right
  
  Const k As Single = 0.1
  Const b As Single = 0.07
  
  With rob(t)
    
  Dim side As Single
  Dim tops As Single
  
  Dim top As Single
  Dim bottom As Single
  Dim Left As Single
  Dim Right As Single
  
  top = .radius + smudgefactor
  bottom = SimOpts.FieldHeight - .radius - smudgefactor
  
  Left = .radius + smudgefactor
  Right = SimOpts.FieldWidth - .radius - smudgefactor
  
  If .pos.x > Right Then whichside = 1
  If .pos.y > bottom Then whichbottom = 1
  
  If .pos.x < Left Then whichside = -1
  If .pos.y < top Then whichbottom = -1
  
  If (whichside <> 0 And whichbottom <> 0) And (Not SimOpts.Dxsxconnected And Not SimOpts.Updnconnected) Then
    'we are in a corner, we need to proceed to the nearest corner instead of the respective edges
    'F += -k(displacement)
    Dim displacement As Single
    Dim corner As vector
    
    If whichside = 1 Then side = Right
    If whichside = -1 Then side = Left
    
    If whichbottom = 1 Then tops = bottom
    If whichbottom = -1 Then tops = top
    
    corner = VectorSet(side, tops)
        
    .ImpulseRes = VectorAdd(.ImpulseRes, VectorScalar(VectorSub(.pos, corner), k))
    .ImpulseRes = VectorAdd(.ImpulseRes, VectorScalar(.vel, b))
    Exit Sub
  End If
  
  
  If whichside <> 0 Then
    If Not SimOpts.Dxsxconnected Then
      If whichside = 1 Then side = Right
      If whichside = -1 Then side = Left
      .ImpulseRes.x = .ImpulseRes.x + -(side - .pos.x) * k + .vel.x * b
    Else
      If whichside = 1 Then
        ReSpawn t, Left, .pos.y
      ElseIf whichside = -1 Then
        ReSpawn t, Right, .pos.y
      End If
    End If
  End If
  
  If whichbottom <> 0 Then
    If Not SimOpts.Updnconnected Then
      If whichbottom = 1 Then tops = bottom
      If whichbottom = -1 Then tops = top
      .ImpulseRes.y = .ImpulseRes.y + -(tops - .pos.y) * k + .vel.y * b
    Else
      If whichbottom = 1 Then
        ReSpawn t, .pos.x, top
      ElseIf whichbottom = -1 Then
        ReSpawn t, .pos.x, bottom
      End If
    End If
  End If
  
  End With
End Sub

'' collisions with borders, or moves robs from one end of
'' the field to the other one, if toroidal is selected
'Public Sub bordercolls2(t As Integer)
'  Dim side As Byte, top As Byte
'
'  With rob(t)
'
'  If .pos.x <= .radius Then
'    If SimOpts.Dxsxconnected = True Then
'      ReSpawn t, SimOpts.FieldWidth - .radius, rob(t).pos.y
'    Else
'      rob(t).mem(214) = 1 'flags collision with edge of screen in mem location 510
'      rob(t).pos.x = .radius
'      rob(t).vel.x = rob(t).vel.x / 2
'    End If
'  End If
'  If rob(t).pos.x >= SimOpts.FieldWidth - .radius Then
'    If SimOpts.Dxsxconnected Then
'      ReSpawn t, .radius, rob(t).pos.y
'    Else
'      rob(t).mem(214) = 1 'flags collision with edge of screen in mem location 510
'      rob(t).pos.x = SimOpts.FieldWidth - .radius
'      rob(t).vel.x = rob(t).vel.x / 2
'    End If
'  End If
'  If rob(t).pos.y <= .radius Then
'    If SimOpts.Updnconnected Then
'      ReSpawn t, rob(t).pos.x, SimOpts.FieldHeight - .radius
'    Else
'      rob(t).mem(214) = 1 'flags collision with edge of screen in mem location 510
'      rob(t).pos.y = .radius
'      rob(t).vel.y = rob(t).vel.y / 2
'    End If
'  End If
'  If rob(t).pos.y >= SimOpts.FieldHeight - .radius Then
'    If SimOpts.Updnconnected Then
'      ReSpawn t, rob(t).pos.x, .radius
'    Else
'      rob(t).mem(214) = 1 'flags collision with edge of screen in mem location 510
'      rob(t).pos.y = SimOpts.FieldHeight - .radius
'      rob(t).vel.y = rob(t).vel.y / 2
'    End If
'  End If
'  t = t
'  End With
'End Sub

Public Sub Colls2(n As Integer)
  Dim k As Integer
  Dim distvector As vector
  Dim dist As Single
  
  With rob(n)
  For k = n + 1 To MaxRobs
    While Not rob(k).Exist
      k = k + 1
      If k > MaxRobs Then Exit Sub
    Wend

    distvector = VectorSub(.pos, rob(k).pos)

    'for an interesting effect, try > instead
    dist = .radius + rob(k).radius
    If VectorMagnitudeSquare(distvector) < dist * dist Then
      Repel2 n, k
    End If
  Next k
  End With
End Sub

'' calculates collisions between the robot pointed by node n
'' (of the robots' sorted linked list) and other robots
'' and generates accelerations
'Public Sub colls(n As node)
'  Dim nd As node
'  Dim t As Integer
'  Dim a As Integer
'  Dim dist As Long
'  Dim tvx As Long, tvy As Long
'  Dim slowdown As Single
'  Dim angl As Single
'  Dim aPer As Single
'  Dim tPer As Single
'  Dim aM As Single
'  Dim tM As Single
'  Dim totM As Single
'  Dim colldist As Long
'  Dim colldistsquared As Long
'  Dim deltax As Long
'  Dim deltay As Long
'  Dim deltaxsquare As Long
'  Dim deltaysquare As Long
'  Dim maxdist As Integer
'  Dim otherrobot As Integer
'  Dim xpos As Long
'
'  slowdown = 0.8
'
'  On Error Resume Next
'
'  t = n.robn
'  xpos = n.xpos
'
'  maxdist = FindRadius(32000) * 2
'
'  Set nd = rlist.firstprox(n, maxdist)
'  otherrobot = rob(nd.robn).order
'
'  If Roborder(otherrobot) = -1 Then Exit Sub
'
'  While rob(Roborder(otherrobot)).pos.x < xpos + maxdist
'    a = Roborder(otherrobot)
'    If t <> a Then
'      colldist = rob(t).radius + rob(a).radius
'      colldistsquared = colldist * colldist 'so we don't need the sqr
'
'      deltax = rob(t).pos.x - rob(a).pos.x
'      If Abs(deltax) > colldist Then GoTo bypass1
'      deltay = rob(t).pos.y - rob(a).pos.y
'      If Abs(deltay) > colldist Then GoTo bypass1
'
'      deltaxsquare = deltax * deltax
'      deltaysquare = deltay * deltay
'      If deltaxsquare + deltaysquare > colldistsquared Then GoTo bypass1
'
'        Repel2 a, t
'        touch t, rob(a).pos.x, rob(a).pos.y
'        touch a, rob(t).pos.x, rob(t).pos.y
'        GoTo bypass
'      'End If
'bypass1:
'
'    'PLEASE don't delete this
'    'this was Numsgil's attempt at physics
'    'there might still be some useful bits here and there
'
'    'deltax = deltax + rob(t).vx - rob(a).vx
'    'If deltax > colldist Then GoTo bypass
'    '
'    'deltay = deltay + rob(t).vy - rob(a).vy
'    'If deltay > colldist Then GoTo bypass
'    '
'    'deltaxsquare = deltax * deltax
'    'deltaysquare = deltay * deltay
'    '
'    'If deltaxsquare + deltaysquare <= colldistsquared Then
'    '
'    '  dist = Sqr(deltaxsquare + deltaysquare)
'    '  If dist = 0 Then GoTo bypass
'    '
'    '  Dim deltaxsingle As Single
'    '  Dim deltaysingle As Single
'    '
'    '  deltaxsingle = deltax
'    '  deltaysingle = deltay
'    '
'    '  deltaxsingle = deltaxsingle / dist
'    '  deltaysingle = deltaysingle / dist
'    '
'    '  Dim tabx As Single
'    '  Dim taby As Single
'    '
'    '  tabx = -deltaysingle
'    '  taby = deltaxsingle
'    '
'    '  Dim vait As Single
'    '  vait = rob(a).vx * tabx + rob(a).vy * taby
'
'    '  Dim vain As Single
'    '  vain = rob(a).vx * deltaxsingle + rob(a).vy * deltaysingle
'
'    '  Dim vbit As Single
'    '  vbit = rob(t).vx * tabx + rob(t).vy * taby
'
'    '  Dim vbin As Single
'    '  vbin = rob(t).vx * deltaxsingle + rob(t).vy * deltaysingle
'
'    '  Dim ma As Single
'    '  ma = rob(a).mass
'    '  If rob(a).Fixed Then ma = 1000000
'
'    '  Dim mb As Single
'    '  mb = rob(t).mass
'    '  If rob(t).Fixed Then mb = 1000000
'
'    '  Dim vafn As Single
'    '  vafn = (mb * vbin * (cof_E + 1) + vain * (ma - cof_E * mb)) / (ma + mb)
'
'    '  Dim vbfn As Single
'    '  vbfn = (ma * vain * (cof_E + 1) - vbin * (ma - cof_E * mb)) / (ma + mb)
'
'    '  Dim vaft As Single
'    '  vaft = vait
'
'    '  Dim vbft As Single
'    '  vbft = vbit
'
'    '  Dim xfa As Single
'    '  xfa = vafn * deltaxsingle + vaft * tabx
'
'    '  Dim yfa As Single
'    '  yfa = vafn * deltaysingle + vaft * taby
'
'    '  Dim xfb As Single
'    '  xfb = vbfn * deltaxsingle + vbft * tabx
'
'    '  Dim yfb As Single
'    '  yfb = vbfn * deltaysingle + vbft * taby
'
'    '  If Sqr(xfa ^ 2 + yfa ^ 2) > 60 Then
'    '    'normalize * maxspeed
'    '  End If
'
'    '  If Sqr(xfb ^ 2 + yfb ^ 2) > 60 Then
'    '    'normalize * maxspeed
'    '  End If
'
'    '  rob(a).vx = xfa
'    '  rob(a).vy = yfa
'
'    '  rob(t).vx = xfb
'    '  rob(t).vy = yfb
'   'End If
'bypass:
'    End If
'    'Set nd = rlist.nextorder(nd)
'    otherrobot = otherrobot + 1
'    If Roborder(otherrobot) = -1 Then
'      Exit Sub
'    End If
'  Wend
'End Sub

'only a minor bug, sometimes (rarely) bots get 'stuck' together, forming an
'infinite energy matrix!  woah!
Public Sub Repel2(rob1 As Integer, rob2 As Integer)
  Dim uv As vector
  Dim vy As vector
  Dim length As Single
  Dim Force As Single
  Dim ForceVector As vector
  
  Const k As Single = 0.1
  Const b As Single = 0.1
  
  uv = VectorSub(rob(rob1).pos, rob(rob2).pos)
  length = VectorInvMagnitude(uv)
                
  If length <> -1# Then 'vectorinvmagnitude = inverse magnitude.  Returns -1# if divide by zero
    uv = VectorScalar(uv, length)
    
    'length is now displacement
    length = rob(rob1).radius + rob(rob2).radius - 1 / length
        
    'Restitutive Force
    Force = k * length
    ForceVector = VectorScalar(uv, Force)
    rob(rob1).ImpulseInd = VectorAdd(rob(rob1).ImpulseInd, ForceVector)
    rob(rob2).ImpulseInd = VectorSub(rob(rob2).ImpulseInd, ForceVector)
      
    'next -bv
    vy = VectorSub(rob(rob2).vel, rob(rob1).vel)
    Force = Dot(vy, uv) * b
    ForceVector = VectorScalar(uv, Force)
    rob(rob1).ImpulseInd = VectorAdd(rob(rob1).ImpulseInd, ForceVector)
    rob(rob2).ImpulseInd = VectorSub(rob(rob2).ImpulseInd, ForceVector)
  End If
End Sub

'' gives to too near robots an accelaration towards
'' opposite directions, inversely prop. to their distance
'Private Sub repel(k As Integer, t As Integer)
'  Dim d As Single
'  Dim dx As Integer
'  Dim dy As Integer
'  Dim dxlk As Single
'  Dim dylk As Single
'  Dim dxlt As Single
'  Dim dylt As Single
'  Dim kconst As Single
'  Dim llink As Single
'  Dim accel As Long  'new acceleration to apply
'  Dim difKE As Long
'
'  Dim maxcel As Single
'  Dim angl1 As Single
'  Dim angl2 As Single
'  Dim colldist As Integer
'  Dim totx As Single 'total x velocity
'  Dim toty As Single 'total y velocity
'  Dim totv As Single 'total absolute velocity
'  Dim totaccel As Single
'  Dim decell As Single
'  Dim kmovx As Single
'  Dim kmovy As Single
'  Dim tmovx As Single
'  Dim tmovy As Single
'  Dim kKE As Single
'  Dim tKE As Single
'  Dim kxm As Single   'robot k x momentum
'  Dim kym As Single   'robot k y momentum
'  Dim txm As Single   'robot t x momentum
'  Dim tym As Single   'robot t y momentum
'  Dim xmean As Single  'Mean of both x momentums
'  Dim ymean As Single  'Mean of both y momentums
'
'  Dim totKE As Single
'  Dim kPer As Single    'percentage of acceleration to give to robot k based on mass
'  Dim tPer As Single    'same for robot t
'  Dim moveaway As Single  'The amount to directly move a robot away from a collision
'
'  'If xmoveaway(k, t) Then GoTo bypass2
'  'If ymoveaway(k, t) Then GoTo bypass2
'
'  dx = (rob(t).pos.x - rob(k).pos.x)
'  dy = (rob(t).pos.y - rob(k).pos.y)
'  colldist = rob(k).Radius + rob(t).Radius 'amount of overlap based on size of robot
'  d = Sqr(dx ^ 2 + dy ^ 2) + 0.01   'inter-robot distance
'
'  decell = 1
'  maxcel = 40
'  'GoTo bypass2
'  kKE = rob(k).mass * rob(k).vel.x + rob(k).mass * rob(k).vel.y
'  tKE = rob(t).mass * rob(t).vel.x + rob(t).mass * rob(t).vel.y
'  kxm = rob(k).mass * rob(k).vel.x     'rob(k) x momentum signed
'  kym = rob(k).mass * rob(k).vel.y     'rob(k) y momentum signed
'  txm = rob(t).mass * rob(t).vel.x     'rob(t) x momentum signed
'  tym = rob(t).mass * rob(t).vel.y     'rob(t) y momentum signed
'  xmean = (Abs(kxm) + Abs(txm)) / 2 'absolute mean of x momentums. Both counted as positive
'  ymean = (Abs(kym) + Abs(tym)) / 2 'absolute mean of y momentums. Both counted as positive
'  If rob(k).mass = 0 Then rob(k).mass = 0.001
'  rob(k).vel.x = rob(k).vel.x + (xmean / rob(k).mass) * kxColdir(k, t) * decell
'  rob(t).vel.x = rob(t).vel.x + (xmean / rob(t).mass) * txColdir(k, t) * decell
'  rob(k).vel.y = rob(k).vel.y + (ymean / rob(k).mass) * kyColdir(k, t) * decell
'  rob(t).vel.y = rob(t).vel.y + (ymean / rob(t).mass) * tyColdir(k, t) * decell
'
'bypass2:
'  'totKE = tKE + kKE   'calculates total momentum of both robots
'  'difKE = kKE - tKE   'difference in momentum
'
'  'tPer = rob(k).mass / (rob(k).mass + rob(t).mass)  'percentage of total momentum in bot k
'  'kPer = rob(t).mass / (rob(k).mass + rob(t).mass)  'percentage of total momentum in bot t
'
'
'
'  'If rob(k).vx = 0 Then rob(k).vx = 0.00001
'  'If rob(t).vx = 0 Then rob(t).vx = 0.00001
'  'mvanglt = Atn(rob(k).vy / rob(k).vx)
'  'mvangla = Atn(rob(t).vy / rob(t).vx)
'  'angledifference = angnorm(AngDiff(mvanglt, angl))
'  'exitangle = angnorm(angl - angledifference)
'  'newvx = (rob(k).vx + rob(t).vx) * (rob(t).mass / (rob(k).mass + rob(t).mass))
'  'newvy = (rob(k).vy + rob(t).vy) * (rob(t).mass / (rob(k).mass + rob(t).mass))
'  'newvx = newvx * 0.95
'  'newvy = newvy * 0.95
'  'llink = 1000
'
'  'totx = Abs(rob(k).vx) + Abs(rob(t).vx)
'  'totx = rob(k).vx + rob(t).vx
'  'toty = Abs(rob(k).vy) + Abs(rob(t).vy)
'  'toty = rob(k).vy + rob(t).vy
'  'totv = totx + toty
'
'
'  angl1 = angle(rob(k).pos.x, rob(k).pos.y, rob(t).pos.x, rob(t).pos.y) 'angle from rob k to rob t
'  angl2 = angle(rob(t).pos.x, rob(t).pos.y, rob(k).pos.x, rob(k).pos.y) 'angle from rob t to rob k
'  'colldist = colldist * 1.2
'
'  'dxlk = absx(angl1, totKE, 0, 0, 0) * tPer
'  'dylk = absy(angl1, totKE, 0, 0, 0) * tPer
'  'dxlt = absx(angl2, totKE, 0, 0, 0) * kPer
'  'dylt = absy(angl2, totKE, 0, 0, 0) * kPer
'  'totaccel = Abs(dxl) + Abs(dyl)
'
'
'  'kconst = 0.01
'  'dxl = (dx - (llink * dx) / d)
'  'dyl = (dy - (llink * dy) / d)
'  moveaway = (colldist - d) / 2 'move away based on half of overlap
'  If d < colldist Then
'    kmovx = absx(angl1, moveaway, 0, 0, 0)
'    kmovy = absy(angl1, moveaway, 0, 0, 0)
'    tmovx = absx(angl2, moveaway, 0, 0, 0)
'    tmovy = absy(angl2, moveaway, 0, 0, 0)
'    If Not rob(t).Fixed Then
'      rob(t).pos.x = rob(t).pos.x - tmovx
'      rob(t).pos.y = rob(t).pos.y - tmovy
'    End If
'    If Not rob(k).Fixed Then
'      rob(k).pos.x = rob(k).pos.x - kmovx
'      rob(k).pos.y = rob(k).pos.y - kmovy
'    End If
'  Else
'    kmovx = 0
'    kmovy = 0
'    tmovx = 0
'    tmovy = 0
'  End If
'bypass3:
'End Sub
'
'Private Function kxColdir(k As Integer, t As Integer)
'  If rob(k).pos.x < rob(t).pos.x Then
'    kxColdir = -1
'  Else
'    kxColdir = 1
'  End If
'End Function
'
'Private Function kyColdir(k As Integer, t As Integer)
'If rob(k).pos.y < rob(t).pos.y Then
'    kyColdir = -1
'  Else
'    kyColdir = 1
'  End If
'End Function
'
'Private Function txColdir(k As Integer, t As Integer)
'If rob(t).pos.x < rob(k).pos.x Then
'    txColdir = -1
'  Else
'    txColdir = 1
'  End If
'End Function
'
'Private Function tyColdir(k As Integer, t As Integer)
'If rob(t).pos.y < rob(k).pos.y Then
'    tyColdir = -1
'  Else
'    tyColdir = 1
'  End If
'End Function
'
'Private Function xmoveaway(k As Integer, t As Integer) As Boolean
'  If Sgn(rob(k).vel.x) = Sgn(rob(t).vel.x) Then 'both moving the same way
'    If rob(k).pos.x < rob(t).pos.x Then           'rob(k) is to the left
'      If Sgn(rob(k).vel.x) = 1 Then          'rob(k) moving to the right
'        If rob(k).vel.x < rob(t).vel.x Then     'rob(k) moving slower than rob(t)
'          xmoveaway = True                'moving away
'        Else
'          xmoveaway = False               'not moving away
'        End If
'      Else                                'rob(k) moving to the left
'        If rob(k).vel.x > rob(t).vel.x Then     'rob(k) moving faster than rob(t)
'          xmoveaway = True                'moving away
'        Else
'          xmoveaway = False               'not moving away
'        End If
'      End If
'    Else                                  'rob(k) is NOT to the left (right or level)
'      If Sgn(rob(k).vel.x) = 1 Then          'rob(k) moving to the right
'        If rob(k).vel.x > rob(t).vel.x Then     'rob(k) moving faster than rob(t)
'          xmoveaway = True                'moving away
'        Else
'          xmoveaway = False               'not moving away
'        End If
'      Else                                'rob(k) moving to the left
'        If rob(k).vel.x < rob(t).vel.x Then     'rob(k) moving faster than rob(t)
'          xmoveaway = True                'moving away
'        Else
'          xmoveaway = False               'not moving away
'        End If
'      End If
'    End If
'  Else                                    'robots moving opposite directions
'    If rob(k).pos.x < rob(t).pos.x Then           'rob(k) is to the left
'      If Sgn(rob(k).vel.x) = 1 Then          'rob(k) moving to the right
'        xmoveaway = False
'      Else                                'rob(k) moving to the left
'        xmoveaway = True
'      End If
'    Else                                  'rob(k) is to the right or level
'      If Sgn(rob(k).vel.x) = 1 Then          'rob(k) moving to the right
'        xmoveaway = True                  'must be moving away
'      Else                                'rob(k) moving to the left
'        xmoveaway = False                 'must be moving towards
'      End If
'    End If
'  End If
'End Function
'
'Private Function ymoveaway(k As Integer, t As Integer) As Boolean
'If Sgn(rob(k).vy) = Sgn(rob(t).vy) Then   'both moving the same way
'    If rob(k).y < rob(t).y Then           'rob(k) is to the top
'      If Sgn(rob(k).vy) = 1 Then          'rob(k) moving to the bottom
'        If rob(k).vy < rob(t).vy Then     'rob(k) moving slower than rob(t)
'          ymoveaway = True                'moving away
'        Else
'          ymoveaway = False               'not moving away
'        End If
'      Else                                'rob(k) moving to the left
'        If rob(k).vy > rob(t).vy Then     'rob(k) moving faster than rob(t)
'          ymoveaway = True                'moving away
'        Else
'          ymoveaway = False               'not moving away
'        End If
'      End If
'    Else                                  'rob(k) is NOT to the left (right or level)
'      If Sgn(rob(k).vy) = 1 Then          'rob(k) moving to the bottom
'        If rob(k).vy > rob(t).vy Then     'rob(k) moving faster than rob(t)
'          ymoveaway = True                'moving away
'        Else
'          ymoveaway = False               'not moving away
'        End If
'      Else                                'rob(k) moving to the top
'        If rob(k).vy < rob(t).vy Then     'rob(k) moving faster than rob(t)
'          ymoveaway = True                'moving away
'        Else
'          ymoveaway = False               'not moving away
'        End If
'      End If
'    End If
'  Else                                    'robots moving opposite directions
'    If rob(k).y < rob(t).y Then           'rob(k) is to the top
'      If Sgn(rob(k).vy) = 1 Then          'rob(k) moving to the bottom
'        ymoveaway = False
'      Else                                'rob(k) moving to the top
'        ymoveaway = True
'      End If
'    Else                                  'rob(k) is to the right or level
'      If Sgn(rob(k).vy) = 1 Then          'rob(k) moving to the bottom
'        ymoveaway = True                  'must be moving away
'      Else                                'rob(k) moving to the top
'        ymoveaway = False                 'must be moving towards
'      End If
'    End If
'  End If
'End Function
